/*
 * Network model with automatic layout
 *
 * Layout code from nem by Elan Amir
 *
 * $Header: /usr/src/mash/repository/vint/nam-1/anetmodel.cc,v 1.6 1997/11/06 06:08:00 haoboy Exp $
 */

#include <stdlib.h>
#include <math.h>

#include "random.h"
#include "view.h"
#include "netview.h"
#include "animation.h"
#include "queue.h"
#include "edge.h"
#include "node.h"
#include "agent.h"
#include "sincos.h"
#include "state.h"
#include "packet.h"
#include "anetmodel.h"

class AutoNetworkModelClass : public TclClass {
public:
	AutoNetworkModelClass() : TclClass("AutoNetworkModel") {}
	TclObject* create(int , const char*const* ) {
		return (new AutoNetModel);
	}
} autonetworkmodel_class;

int AutoNetModel::AUTO_ITERATIONS_ = 1;
int AutoNetModel::INCR_ITERATIONS_ = 200;
int AutoNetModel::RANDOM_SEED_ = 1;
double AutoNetModel::INIT_TEMP_ = 0.10; // follow Elan's 
double AutoNetModel::MINX_ = 0.0; 
double AutoNetModel::MAXX_ = 1.0;
double AutoNetModel::MINY_ = 0.0;
double AutoNetModel::MAXY_ = 1.0;

AutoNetModel::AutoNetModel() : NetModel()
{
	iterations_ = AUTO_ITERATIONS_;
	bind("KCa_", &kca_);
	bind("KCr_", &kcr_);
	bind("INCR_ITERATIONS_", &INCR_ITERATIONS_);
	bind("AUTO_ITERATIONS_", &AUTO_ITERATIONS_);
	bind("RANDOM_SEED_", &RANDOM_SEED_);
}

AutoNetModel::~AutoNetModel()
{
}

int AutoNetModel::command(int argc, const char *const *argv)
{
	if (argc == 2) {
		if (strcmp(argv[1], "layout") == 0) {
			/*
			 * <net> layout
			 * compute layout using Elan's nem
			 */
			layout();
			return (TCL_OK);
		}
		if (strcmp(argv[1], "relayout") == 0) {
			relayout();
			return (TCL_OK);
		}
	} 
	return (NetModel::command(argc, argv));
}

void AutoNetModel::initEmbedding() 
{
	Node *n;
	double maxx, maxy;

	Random::seed(RANDOM_SEED_);

	maxx = MINX_ + (MAXX_-MINX_) * 0.4;
	maxy = MINY_ + (MAXY_-MINY_) * 0.4;	

	// randomize nodes' position within square [(0,0), (1,1)]
	for (nNodes_ = 0, n = nodes_; n != 0; n = n->next_, nNodes_++) {
		n->place(Random::uniform(MINX_, maxx),
			 Random::uniform(MINY_, maxy));
#if 0
		printf("Node %s: initial position(%f, %f)\n",
		       n->name(), n->x(), n->y());
#endif
	}
	//optk_ = kc_ * sqrt((MAXX_-MINX_)*(MAXY_-MINY_) / (double)nNodes);
	optka_ = kca_ * sqrt((MAXX_-MINX_)*(MAXY_-MINY_) / (double)nNodes_);
	optkr_ = kcr_ * sqrt((MAXX_-MINX_)*(MAXY_-MINY_) / (double)nNodes_);
	temp_ = INIT_TEMP_;
}

void AutoNetModel::placeEverything()
{
	for (Node *n = nodes_; n != 0; n = n->next_) {
		for (Edge *e = n->links(); e != 0; e = e->next_)
			placeEdge(e, n);
		placeAllAgents(n);
		if (nymin_ > n->y())
			nymin_ = n->x();
		if (nymax_ < n->y())
			nymax_ = n->y();
	}
}

// do more passes
void AutoNetModel::relayout()
{
	Node *n;
	Edge *e;

	for (n = nodes_; n != 0; n = n->next_)
		for (e = n->links(); e != 0; e = e->next_)
			e->unmark();

	// in case that the two constants changed
	optka_ = kca_ * sqrt((MAXX_-MINX_)*(MAXY_-MINY_) / (double)nNodes_);
	optkr_ = kcr_ * sqrt((MAXX_-MINX_)*(MAXY_-MINY_) / (double)nNodes_);

	for (int i = 0; i < 100; i++)
		embed();

	scale_estimate();
	placeEverything();

	for (NetView *p = views_; p != 0; p = p->next_)
		p->redrawModel();
}

// do several passes of embedding
void AutoNetModel::layout()
{
	initEmbedding();
	for (int i = 0; i < iterations_; i++) {
		embed();
	}

	scale_estimate(); // find node size after layout
	placeEverything();
}

void AutoNetModel::cool()
{
	if (temp_ > 0.001)
		temp_ *= 0.95;
	else 
		temp_ = 0.001;
}

void AutoNetModel::placeAgent(Agent *a, Node *src) const
{
	double x0, y0, nsin, ncos;
	if (a->marked() == 0) {
		if (a->angle()==NO_ANGLE) {
			if (a->edge()==NULL) {
				a->angle(0);
			} else {
				a->angle(a->edge()->realangle()+M_PI/4.0);
			}
		}
		SINCOS(a->angle(), &nsin, &ncos);
		x0 = src->x() + src->size() * ncos * .75;
		y0 = src->y() + src->size() * nsin * .75;
		a->place(x0, y0);
		a->mark(1);
	}
}

void AutoNetModel::placeAllAgents(Node *src) const
{
	// clear all marks and clear all angles
	for (Agent *a = src->agents(); a != NULL; a = a->next()) {
		a->mark(0), a->angle(NO_ANGLE);
		placeAgent(a, src);
	}
}

// we need to place edge using the edge's real angle, instead of the 
// given one
void AutoNetModel::placeEdge(Edge* e, Node* src) const
{
	if (e->marked() == 0) {
		double nsin, ncos, angle;
		Node *dst = e->neighbor();
		angle = src->angle(*dst);
		SINCOS(angle, &nsin, &ncos);
		double x0 = src->x() + src->size() * ncos * .75;
		double y0 = src->y() + src->size() * nsin * .75;
		double x1 = dst->x() - dst->size() * ncos * .75;
		double y1 = dst->y() - dst->size() * nsin * .75;
		e->place(x0, y0, x1, y1);
		
		/* Place the queue here too.  */
		EdgeHashNode *h = lookupEdge(e->src(), e->dst());
		if (h->queue != 0)
			h->queue->place(e->size(), e->x0(), e->y0());

		e->mark();
	}
}

// we need to use edge length, instead of delay, to estimate scale
void AutoNetModel::scale_estimate()
{
	/* Determine the maximum link delay. */
	double emax = 0., dist;
	Node *n;
	for (n = nodes_; n != 0; n = n->next_) {
		for (Edge* e = n->links(); e != 0; e = e->next_) {
			dist = n->distance(*(e->neighbor()));
			if (dist > emax)
				emax = dist;
		}
	}
	
	/*
	 * Check for missing node or edge sizes. If any are found,
	 * compute a reasonable default based on the maximum edge
	 * dimension.
	 */
	for (n = nodes_; n != 0; n = n->next_) {
		n->size(.1 * emax);
		for (Edge* e = n->links(); e != 0; e = e->next_)
			e->size(.03 * emax);
	}
	/*store this because we need it for monitors*/
	node_size_=.1 * emax;
}

// Fruchterman, T.M.J and Reingold, E.M.,
// Graph Drawing by Force-directed Placement,
// Software-Practice and Experience, vol. 21(11), 1129-1164 (Nov. 91)
//
// Do one pass of embedding
void AutoNetModel::embed()
{
	Node *v, *u;	
	Edge* e;
	double dx, dy, mag, f, minn;

	/*
	 * Calculate repulsive forces between all vertices.
	 */
	for (v = nodes_; v != 0; v = v->next_) {
		//XXX why not 0. as in paper?
		v->displace(0., 0.);
		for (u = nodes_; u != 0; u = u->next_) {
			if (u == v) 
				continue;
			dx = v->x() - u->x();
			dy = v->y() - u->y();
			if (dx == 0 && dy == 0)
				dx = dy = 0.001;
			mag = sqrt(dx * dx + dy * dy);
			f = REP(mag, optkr_);
			v->displace(v->dx() + dx / mag * f,
				    v->dy() + dy / mag * f);
		}
	}

	/*
	 * Calculate attractive forces between neighbors.
	 */
	for (v = nodes_; v != 0; v = v->next_) {
		for (e = v->links(); e != 0; e = e->next_) {
			u = e->neighbor();
			dx = v->x() - u->x();
			dy = v->y() - u->y();
			mag = sqrt(dx * dx + dy * dy);

			// XXX we consider single direction edge as 
			// of single direction attraction. So we only
			// attract one node toward the other. If later 
			// we have the other edge, the other node will be 
			// attracted too.
			if (mag != 0) {
				f = ATT(mag, optka_);
				v->displace(v->dx() - dx / mag * f,
					    v->dy() - dy / mag * f);
			}
		}
	}

	/* 
	 * Limit the maximum displacement to the temperature temp;
	 * and then prevent from being displaced outside frame.
	 */
	for (v = nodes_; v != 0; v = v->next_) {
		mag = sqrt(v->dx()*v->dx() + v->dy() * v->dy());
		double posx = v->x(), posy = v->y();
		if (mag != 0) {
			minn = min(mag, temp_);
			posx += v->dx() / mag * minn;
			posy += v->dy() / mag * minn;
		}

#if 0
		posx = min(MAXX_, max(MINX_, posx));
		posy = min(MAXY_, max(MINY_, posy));
#endif

		v->place(posx, posy);

#if 0
		printf("Position of node %s: (%f, %f)\n", 
		       v->name(), posx, posy);
#endif
	}

	/* Cool the temperature */
	cool();

#if 0
	printf("------------------------------\n");
#endif
}


// packet handling stuff. use new packet
void AutoNetModel::handle(const TraceEvent& e, double now, int direction)
{
	switch (e.tt) {
	case 'a':
		NetModel::handle(e, now, direction);
		// recalculate bounding box so that all agents will be within
		// view
		for (NetView *v = views_; v != 0; v = v->next_)
			v->redrawModel();
		break;
	default:
		NetModel::handle(e, now, direction);
		break;
	}
}


